/*
 * mnemnonic.c - Ermoegliche es Motif-Applikationen, auch PushButtons
 *               und dergleichen mittels Kurztastenkombination auszu-
 *               loesen, wenn diese Widgets sich nicht in einem Menu
 *               befinden.
 * Letzte Modifikation: 1.3.1994
 * Harald Albrecht ( albrecht@igpm.rwth-aachen.de )
 */
#include <X11/keysym.h>
#include <Xm/Xm.h>
#include <Xm/Label.h>
#include <Xm/RowColumn.h>
/* --------------------------------------------------------------------------
 * Waehle das gefundene Widget an (gib' ihn also den Tastaturfokus)
 * und schicke ihm dann einen Tastendruck, um es auf diese Weise zu
 * aktivieren.
 */
static void ActivateWidget(Widget w, XKeyEvent *pKeyEvent)
{
    XKeyEvent  KeyEvent;
    Display    *display;
    Window     window;

    if ( !XmProcessTraversal(w, XmTRAVERSE_CURRENT) ) return;
    display            = XtDisplayOfObject(w);
    window             = XtWindowOfObject(w);
    KeyEvent           = *pKeyEvent;
    KeyEvent.type      = KeyPress;
    KeyEvent.window    = window;
    KeyEvent.subwindow = window;
    KeyEvent.state     = 0;
    KeyEvent.keycode   = XKeysymToKeycode(display, XK_space);
    XtDispatchEvent((XEvent *) &KeyEvent);
    KeyEvent.type      = KeyRelease;
    XtDispatchEvent((XEvent *) &KeyEvent);
} /* ActivateWidget */
/* --------------------------------------------------------------------------
 * Einfache Hilfsfunktion, die angibt, ob ein fragliches Widget auch
 * generell bereit waere, den Tastaturfokus anzunehmen.
 */
Boolean XtTraversalOn(Widget w)
{
    Boolean TraversalOn = False;
    
    if ( XtIsSubclass(w, xmPrimitiveWidgetClass) ||
         XtIsSubclass(w, xmManagerWidgetClass) )
	XtVaGetValues(w, XmNtraversalOn, &TraversalOn, NULL);
    return TraversalOn;
} /* XtTraversalOn */
/* --------------------------------------------------------------------------
 * Mit einem gegebenen Widget suche nun nach einem Empfaenger innerhalb
 * dieser angegebenen Hierarchie. Ist es ein Nachkomme von XmPrimitive, 
 * so bekommt das Widget sofort den Fokus, kein Problem. Ist es hingegen
 * ein Nachkomme von XmManager, so muessen wir das erste Kind suchen, dass
 * den Fokus annehmen kann. Sonderfall: sollte eines der Kinder dieses
 * Managers bereits den Fokus haben, lassen wir ihm den Fokus weiterhin.
 */
#if defined(__cplusplus) || defined(c_plusplus)
extern "C" Widget _XmGetTabGroup(Widget w);
#else
extern Widget _XmGetTabGroup(Widget w);
#endif
static void FindHomeWidget(Widget w, XKeyEvent *pKeyEvent)
{
    int         i, NumChildren;
    WidgetList  Children;
    Widget      Child;
    Boolean     ManagerTraversed = False;

    while ( XtIsSubclass(w, xmManagerWidgetClass) ) {
        ManagerTraversed = True;
	XtVaGetValues(w, XmNchildren, &Children, 
			 XmNnumChildren, &NumChildren, NULL);
	for ( i = 0; i < NumChildren; ++i ) {
	    Child = Children[i];
	    if ( XtIsManaged(Child) && XtIsRealized(Child) &&
	         XtIsSensitive(Child) && XtTraversalOn(Child) ) 
		{ w = Child; break; }
	}
	if ( i == NumChildren ) return; /* nix gefunden */
    }
    if ( ManagerTraversed &&
         (_XmGetTabGroup(w) == XtParent(w)) ) return;
    XmProcessTraversal(w, XmTRAVERSE_CURRENT);
    if ( ManagerTraversed ) XmProcessTraversal(w, XmTRAVERSE_HOME);
} /* FindHomeWidget */
/* --------------------------------------------------------------------------
 * Suche ein geeignetes Widget, dass mit groesster Wahrscheinlichkeit
 * zu dem Label-Widget gehoert, welches den soeben vom Benutzer ein-
 * gegebenen Mnemonic enthaelt.
 */
static void GiveFocusToAnotherWidget(Widget w, XKeyEvent *pKeyEvent)
{
    Boolean       LToR, HTraversalOn, VTraversalOn;
    Position      ChildLeft, ChildRight, ChildTop, ChildBottom;
    Dimension	  ChildWidth, ChildHeight;
    Position      LabelLeft, LabelRight, LabelTop, LabelBottom;
    Dimension	  LabelWidth, LabelHeight;
    Position      PosX, PosY, Left, Right, Top, Bottom;
    WidgetList    Children;
    int           NumChildren, i;
    Widget        HWidget, VWidget;
    Dimension     HGoodness, VGoodness, Goodness;
    Dimension     HDist, VDist, Dist;
    unsigned char StringDirection;
    
    XtVaGetValues(w, XmNstringDirection, &StringDirection, NULL);
    LToR = !(StringDirection != XmSTRING_DIRECTION_L_TO_R);
    LabelTop = 0; LabelLeft = 0;
    XtVaGetValues(w, XmNwidth, &LabelWidth, 
                     XmNheight, &LabelHeight, NULL);
    /*
     * Gehe solange in der Hierarchie nach oben, wie das Label in
     * einem Widget steckt, dass seinerseits keine weiteren Kinder
     * hat. Beim ersten Elternteil, dass auch noch andere Kinder
     * hat, muss genauer gesucht werden.
     */
    for (;;) {
        XtVaGetValues(w, XmNx, &PosX, XmNy, &PosY, NULL);
	LabelLeft += PosX; LabelTop += PosY;
	w = XtParent(w);
	XtVaGetValues(w, XmNchildren, &Children, 
	                 XmNnumChildren, &NumChildren, NULL);
	if ( NumChildren > 1 ) break;
	if ( XtIsSubclass(w, shellWidgetClass) ) return;
    }
    LabelRight  = LabelLeft + LabelWidth  - 1;
    LabelBottom = LabelTop  + LabelHeight - 1;
    /* 
     * Ein moegliches Ziel eines Labelwidgets mit Mnemonic darf sich nur
     * in demjenigen Zweig befinden, der zum dem Label am naechsten
     * liegenden XmManager gehoert. Suche hier nun nach dem besten
     * Kandidaten einmal in horizontaler und einmal in vertikaler
     * Richtung und bewerte dann die Chancen der beiden Kontrahenten.
     */
    HGoodness = 0; HDist = ~0; HWidget = NULL;
    VGoodness = 0; VDist = ~0; VWidget = NULL;
    for ( i = 0; i < NumChildren; ++i ) {
	if ( !XtIsManaged(Children[i]) ||
	     XtIsSubclass(Children[i], shellWidgetClass) ) continue;
	XtVaGetValues(Children[i], XmNx, &ChildLeft, XmNy, &ChildTop, 
	                           XmNwidth, &ChildWidth, 
				   XmNheight, &ChildHeight, NULL);
	ChildRight  = ChildLeft + ChildWidth  - 1;
	ChildHeight = ChildTop  + ChildHeight - 1;
	if ( ( LToR && (ChildLeft > LabelRight) ||
	      !LToR && (ChildRight < LabelLeft) ) &&
	     (ChildTop <= LabelBottom) && (ChildBottom >= LabelTop) ) {
	    Top      = ChildTop    > LabelTop    ? ChildTop    : LabelTop;
	    Bottom   = ChildBottom < LabelBottom ? ChildBottom : LabelBottom;
	    Goodness = Bottom - Top + 1;
	    Dist     = LToR ? ChildLeft - LabelRight : LabelLeft - ChildRight;
	    if (  (Goodness >  HGoodness) ||
	         ((Goodness == HGoodness) && (Dist < HDist)) ) {
		HGoodness = Goodness; HDist = Dist;
		HWidget = Children[i];
	    }
	} else if ( (ChildTop > LabelBottom) &&
	     (ChildLeft <= LabelRight) && (ChildRight >= LabelLeft) ) {
	    Left     = ChildLeft  > LabelLeft  ? ChildLeft  : LabelLeft;
	    Right    = ChildRight < LabelRight ? ChildRight : LabelRight;
	    Goodness = Right - Left + 1;
	    Dist     = ChildTop - LabelBottom;
	    if (  (Goodness >  VGoodness) ||
	         ((Goodness == VGoodness) && (Dist < VDist)) ) {
		VGoodness = Goodness; VDist = Dist;
		VWidget = Children[i];
	    }
	}
    } /* for */
    /*
     * Ok. Entscheide dich nun fuer das besser geeignet erscheinende
     * Widget und ueberlasse ihm den Tastaturfokus.
     */
    HTraversalOn = HWidget != NULL ? XtTraversalOn(HWidget) : False;
    VTraversalOn = VWidget != NULL ? XtTraversalOn(VWidget) : False;
    if ( VTraversalOn && ((HDist > VDist) || !HTraversalOn) ) {
	FindHomeWidget(VWidget, pKeyEvent);
    } else if ( HTraversalOn && ((HDist <= VDist) || !VTraversalOn) ) {
	FindHomeWidget(HWidget, pKeyEvent);
    }
} /* GiveFocusToAnotherWidget */
/* --------------------------------------------------------------------------
 * Rekursiv den Widget-Baum durchkaemmen und dabei nach einem geeigneten
 * Label-Nachkommen suchen, der auf den angegebenen Mnemonic 'hoert'.
 * Parameter:
 *   w		    Aktuelles Widget, das daraufhin zu ueberpruefen ist, 
 *		    ob es ein Empfaenger des gesuchten Tastendruckes ist
 *   LowerKeySym    zu suchende Kurztastenkombination (ohne Shift)
 *   UpperKeySym    dto, aber mit Shift
 *   pKeyEvent      Zeiger auf dasjenige Tastaturereignis, welches diese
 *		    Suche dereinst ausloeste.
 * Ergebnis:
 *   True, wenn der Tastendruck an einen geeigneten Empfaenger ausge-
 *   liefert werden konnte.
 */
static Boolean TraverseWidgetTree(Widget w, 
                                  KeySym LowerKeySym, 
				  KeySym UpperKeySym, 
                                  XKeyEvent *pKeyEvent)
{
    WidgetList ChildList;
    int        NumChildren, Child;
    KeySym     LabelMnemonic;
    
    if ( !XtIsManaged(w) || !XtIsRealized(w) ) return False;
    /*
     * Zuerst einmal ueberpruefen, ob das aktuell zu untersuchende
     * Widget ein Nachkomme von XmLabel ist, da nur dann auch die
     * Mnemonic-Resource existiert.
     */
    if ( XtIsSubclass(w, xmLabelWidgetClass) ) {
        XtVaGetValues(w, XmNmnemonic, &LabelMnemonic, NULL);
        if ( (LabelMnemonic == LowerKeySym) ||
	     (LabelMnemonic == UpperKeySym) ) {
	    if ( !XtIsSensitive(w) ) return True;
	    if ( XtClass(w) == xmLabelWidgetClass )
		 GiveFocusToAnotherWidget(w, pKeyEvent);
	    else ActivateWidget(w, pKeyEvent);
	    return True;
        }
    }
    /*
     * Falls es ein Nachkomme von Composite ist, muessen alle dessen
     * Kinder abgesucht werden: der klassische Fall fuer Rekursion.
     */
    if ( XtIsSubclass(w, compositeWidgetClass) ) {
        if ( XtIsSubclass(w, xmRowColumnWidgetClass) ) {
            /*
	     * Sollten wir soeben auf eine Menuezeile 'getreten' sein,
             * so sollten wir tunlichst davon die Finger lassen und das
             * Feld lieber Motif ueberlassen...!
             */
            unsigned char RowColumnType;
            XtVaGetValues(w, XmNrowColumnType, &RowColumnType, NULL);
            if ( RowColumnType != XmWORK_AREA ) return False;
        } else if ( XtIsSubclass(w, shellWidgetClass) )
	    /*
	     * ebenso nicht in einen Dialog oder dergleichen hinein-
	     * laufen, denn ab einer Shell beginnt ein neuer Ast, der
	     * uns nichts zu interessieren hat.
	     */
	    return False;
        XtVaGetValues(w, XmNchildren,    &ChildList, 
                         XmNnumChildren, &NumChildren, NULL);
        for ( Child = 0; Child < NumChildren; ++Child )
            if ( TraverseWidgetTree(ChildList[Child], 
	                            LowerKeySym, UpperKeySym, 
				    pKeyEvent) )
                return True;
    }
    return False;
} /* TraverseWidgetTree */
/* --------------------------------------------------------------------------
 * Falls ein Tastendruck der Form META+Taste auftauchte, genauer ueber-
 * pruefen, ob es ein moeglicher Mnemonic sein kann und danach erst die
 * Suche nach einem geeigneten Widget beginnen.
 * Parameter:
 *   w		Das Widget, das den Beginn eines zu durchsuchenden
 *		Astes in der Widgethierarchie kennzeichnet.
 *   Event	Benachrichtigung ueber einen Tastendruck, der evtl.
 *		einen Mnemonic darstellt und nach dem ggf. gesucht
 *		werden soll.
 */
static Boolean HandleMnemonic(Widget w, XKeyEvent Event)
{
           Widget         widget, oldWidget;
           int            CharCount;
    static char           KeyBuffer[10], MnemonicBuffer[10];
    static XComposeStatus ComposeStatus;
           KeySym         LowerKey, UpperKey, Keysym;
    
    CharCount = XLookupString(&Event, KeyBuffer, sizeof(KeyBuffer)-1,
                              &Keysym, &ComposeStatus);
    /* 
     * Es darf nur ein einzelnes Zeichen dabei herauskommen, denn sonst
     * kann es kein Mnemonic sein. Suche dann nach dem obersten Shell-
     * Widget und starte von diesem aus die Suche nach einem geeigneten
     * Button... Naja starte lieber mit dem Kind der Shell, denn die
     * Shell kann ja keinen Mnemonic besitzen!
     */
    if ( CharCount == 1 ) {
	XtConvertCase(XtDisplayOfObject(w), Keysym, &LowerKey, &UpperKey);
        widget = w; oldWidget = w;
        while ( (widget != NULL) &&
                !XtIsSubclass(widget, shellWidgetClass) ) {
            oldWidget = widget; widget = XtParent(widget);
        }
	widget = oldWidget;
        return TraverseWidgetTree(widget, LowerKey, UpperKey, &Event);
    }
    return False;
} /* HandleMnemonic */
/* --------------------------------------------------------------------------
 * Dieses ist der Ersatz fuer die bisherige Funktion in der libXt.a
 * Hier wird die (Sonder-) Behandlung von Tastendruecken der Form
 * META+Buchstabe behandelt.
 */
void XtAppMainLoop(XtAppContext AppContext)
{
    XEvent NextEvent;
    Widget w;
    
    for (;;) {
        XtAppNextEvent(AppContext, &NextEvent);
        switch ( NextEvent.type ) {
        case KeyPress:
            if ( NextEvent.xkey.state & Mod1Mask ) {
                w = XtWindowToWidget(NextEvent.xany.display,
                                     NextEvent.xany.window);
                if ( w != NULL && 
		    HandleMnemonic(w, NextEvent.xkey) )
                    break; /* nicht weiterleiten, da bereits bearbeitet. */
            }
        default:
            XtDispatchEvent(&NextEvent);
        }
    }
} /* XtAppMainLoop */

/* Ende von mnemonic.c */
